# Python file to read a XTF file 
# Based on gFXtfReadHeader.m and gFXtfRead000.m by Ivan V. Dmitriev
# All variable names and comments from M-files
# Faro, Seg Set 13 13:59:50 WEST 2021
# Written by Orlando Camargo Rodriguez

#======================================================================
# Triton Imaging, Inc.
# eXtended Triton Format (XTF) Rev. 34
# BYTE   => 1 byte  => Unsigned integer B
# WORD   => 2 bytes => Unsigned integer h
# DWORD  => 4 bytes => Unsigned integer I
# float  => 4 bytes => float f
# double => 8 bytes => float d
#======================================================================

import os
from numpy import *
from matplotlib.pyplot import *

fName = "sonar_data110520154300.xtf"
fSize = os.path.getsize( fName ); # print fSize

print("Reading XTF file...")

fId = open(fName, "rb")
# ===Begin XTFFILEHEADER Structure Read
HFileFormat                     = int( fromfile( fId, int8, 1 ) ) # Should be equal to 123 (0x7B)
HSystemType                     = int( fromfile( fId, int8, 1 ) ) # Set to 1
HRecordingProgramName           = fId.read(8)  # Example: "Isis" or "KleinXlt"
HRecordingProgramVersion        = fId.read(8)  # Example: "556" for version 5.56
HSonarName                      = fId.read(16) # Name of server used to access sonar. Example: "C31_SERV.EXE"
HSonarType                      = int( fromfile( fId, int16, 1 ) )
HNoteString                     = fId.read(64) # Notes as entered in the Sonar Setup dialog box
HThisFileName                   = fId.read(64) # Name of this file. Example:"LINE12-B.XTF"
HNavUnits                       = int(   fromfile( fId, int16  , 1 ) ) # 0 = Meters (i.e., UTm] or 3 = Lat/Long
HNumberOfSonarChannels          = int(   fromfile( fId, int16  , 1 ) ) # if > 6, header grows to 2K in size
HNumberOfBathymetryChannels     = int(   fromfile( fId, int16  , 1 ) )
HNumberOfSnippetChannels        = int(   fromfile( fId, int8   , 1 ) )
HNumberOfForwardLookArrays      = int(   fromfile( fId, int8   , 1 ) )
HNumberOfEchoStrengthChannels   = int(   fromfile( fId, int16  , 1 ) )
HNumberOfInterferometryChannels = int(   fromfile( fId, int8   , 1 ) )
HReserved1                      = int(   fromfile( fId, int8   , 1 ) ) # Reserved. Set to 0.
HReserved2                      = int(   fromfile( fId, int16  , 1 ) ) # Reserved. Set to 0.
HReferencePointHeigh            = float( fromfile( fId, float32, 1 ) ) # Height of reference point above water line [m] 
HProjectionType                 = fId.read(12) # Navigation System Parameters. Not currently used. Set to 0.
HSpheriodType                   = fId.read(10) # Navigation System Parameters. Not currently used. Set to 0.
HNavigationLatency              = int(   fromfile( fId, int32  , 1 ) ) 
# Navigation System Parameters. Latency of nav system in milliseconds. (Usually GPS). 
# ISIS Note: This value is entered on the Serial port setup dialog box.  
# When computing a position, Isis will take the time of the navigation and subtract this value.
HOriginY                        = float( fromfile( fId, float32, 1 ) ) # Navigation System Parameters. Not currently used. Set to 0
HOriginX                        = float( fromfile( fId, float32, 1 ) ) # Navigation System Parameters. Not currently used. Set to 0
HNavOffsetY                     = float( fromfile( fId, float32, 1 ) ) 
# Navigation System Parameters. Orientation of positive Y is forward. 
# ISIS Note: This offset is entered in the Multibeam setup dialog box
HNavOffsetX                     = float( fromfile( fId, float32, 1 ) ) 
# Navigation System Parameters. Orientation of positive X is to starboard. 
# ISIS Note: This offset is entered in the Multibeam setup dialog box
HNavOffsetZ                     = float( fromfile( fId, float32, 1 ) )
# Navigation System Parameters. Orientation of positive Z is down. Just like depth. 
# ISIS Note: This offset is entered in the Multibeam setup dialog box
HNavOffsetYaw                   = float( fromfile( fId, float32, 1 ) ) 
# Navigation System Parameters. Orientation of positive yaw is turn to right.
# ISIS Note: This offset is entered in the Multibeam setup dialog box
HMRUOffsetY                     = float( fromfile( fId, float32, 1 ) ) 
# Navigation System Parameters. Orientation of positive Y is forward.
# ISIS Note: This offset is entered in the Multibeam setup dialog box
HMRUOffsetX                     = float( fromfile( fId, float32, 1 ) ) 
# Navigation System Parameters. Orientation of positive X is to starboard.
# ISIS Note: This offset is entered in the Multibeam setup dialog box
HMRUOffsetZ              = float( fromfile( fId, float32, 1 ) )
# Navigation System Parameters. Orientation of positive Z is down. Just like depth. 
# ISIS Note: This offset is entered in the Multibeam setup dialog box
HMRUOffsetYaw            = float( fromfile( fId, float32, 1 ) ) 
# Navigation System Parameters. Orientation of positive yaw is turn to right.
# ISIS Note: This offset is entered in the Multibeam setup dialog box
HMRUOffsetPitch          = float( fromfile( fId, float32, 1 ) ) 
# Navigation System Parameters. Orientation of positive pitch is nose up.
# ISIS Note: This offset is entered in the Multibeam setup dialog box.
HMRUOffsetRoll           = float( fromfile( fId, float32, 1 ) ) 
# Navigation System Parameters. Orientation of positive roll is lean to starboard. 
# ISIS Note: This offset is entered in the Multibeam setup dialog box
nChanel = HNumberOfSonarChannels + HNumberOfBathymetryChannels + HNumberOfSnippetChannels + HNumberOfForwardLookArrays + HNumberOfEchoStrengthChannels + HNumberOfInterferometryChannels
#===End XTFFILEHEADER Structure Read

#===Begin CHANINFO Structure Allocate
CTypeOfChannel    = zeros(nChanel)
CSubChannelNumber = zeros(nChanel)
CCorrectionFlags  = zeros(nChanel)
CUniPolar         = zeros(nChanel)
CBytesPerSample   = zeros(nChanel)
CReserved         = zeros(nChanel)
CChannelName      = chararray((16,nChanel))
CVoltScale        = zeros(nChanel)
CFrequency        = zeros(nChanel)
CHorizBeamAngle   = zeros(nChanel)
CTiltAngle        = zeros(nChanel)
CBeamWidth        = zeros(nChanel)
COffsetX          = zeros(nChanel)
COffsetY          = zeros(nChanel)
COffsetZ          = zeros(nChanel)
COffsetYaw        = zeros(nChanel)
COffsetPitch      = zeros(nChanel)
COffsetRoll       = zeros(nChanel)
CBeamsPerArray    = zeros(nChanel)
CReservedArea2    = chararray((54,nChanel))
#===End CHANINFO Structure Allocate

for m in range(nChanel):
    #===Begin CHANINFO Structure Read
    CTypeOfChannel[   m] = int(   fromfile( fId, int8   , 1 ) ) # SUBBOTTOM=0, PORT=1, STBD=2, BATHYMETRY=3
    CSubChannelNumber[m] = int(   fromfile( fId, int8   , 1 ) ) # Index for which CHANINFO structure this is
    CCorrectionFlags[ m] = int(   fromfile( fId, int16  , 1 ) ) 
    # 1 = sonar imagery stored as slant-range, 2 = sonar imagery stored as ground range (corrected)
    CUniPolar[        m] = int(   fromfile( fId, int16  , 1 ) ) # 0 = data is polar, 1 = data is unipolar
    CBytesPerSample[  m] = int(   fromfile( fId, int16  , 1 ) ) # 1 (8-bit data) or 2 (16-bit data) or 4 (32-bit)
    CReserved[        m] = int(   fromfile( fId, int32  , 1 ) ) 
# Isis Note: Previously this was SamplesPerChannel.
# Isis now supports the recording of every sample per ping, which means that number of samples per channel can vary from ping to ping 
# if the range scale changes. Because of this, the NumSamples value in the XTFPINGCHANHEADER structure (defined in Section 3.18) holds 
# the number of samples to read for a given channel. 
# For standard analog systems, this Reserved value is still filled in with 1024, 2048 
# or whatever the initial value is for SamplesPerChannel.
    CChannelName[   :,m] = fId.read(16) # Text describing channel.  i.e., "Port 500"
    CVoltScale[       m] = float( fromfile( fId, float32, 1 ) )  
# This states how many volts are represented by a maximum sample value in the range  [-5.0 to +4.9998] volts. Default is 5.0.
    CFrequency[       m] = float( fromfile( fId, float32, 1 ) ) # Center transmit frequency
    CHorizBeamAngle[  m] = float( fromfile( fId, float32, 1 ) ) # Typically 1 degree or so
    CTiltAngle[       m] = float( fromfile( fId, float32, 1 ) ) # Typically 30 degrees
    CBeamWidth[       m] = float( fromfile( fId, float32, 1 ) ) # 3dB beam width, Typically 50 degrees
    COffsetX[         m] = float( fromfile( fId, float32, 1 ) ) # Orientation of positive X is to starboard. 
                                                                # Note: This offset is entered in the Multibeam setup dialog box
    COffsetY[         m] = float( fromfile( fId, float32, 1 ) ) # Orientation of positive Y is forward. 
                                                                # Note: This offset is entered in the Multibeam setup dialog box
    COffsetZ[         m] = float( fromfile( fId, float32, 1 ) ) # Orientation of positive Z is down. Just like depth. 
                                                                # Note: This offset is entered in the Multibeam setup dialog box
    COffsetYaw[       m] = float( fromfile( fId, float32, 1 ) ) # Orientation of positive yaw is turn to right. 
                                                                # If the multibeam sensor is reverse mounted (facing backwards), 
                                                                # then OffsetYaw will be around 180 degrees.
                                                                # Note: This offset is entered in the Multibeam setup dialog box
    COffsetPitch[     m] = float( fromfile( fId, float32, 1 ) ) # Orientation of positive pitch is nose up. 
                                                                # Note: This offset is entered in the Multibeam setup dialog box
    COffsetRoll[      m] = float( fromfile( fId, float32, 1 ) ) # Orientation of positive roll is lean to starboard. 
                                                                # Note: This offset is entered in the Multibeam setup dialog box
    CBeamsPerArray[   m] = int(   fromfile( fId, int16  , 1 ) ) # For forward look only (i.e., Sonatech DDS)
    CReservedArea2[ :,m] = fId.read(54) # Unused Set value to 0
    #===End CHANINFO Structure Read

i1 = fId.tell()
i2 = 1024-(i1%1024)
fId.seek(i2,1)
ss = fId.tell()
nRec = 0

# ===Begin Calc Num of ShortHeader structure
while fSize > fId.tell():
    nRec = nRec + 1
    face = int( fromfile( fId, int16, 1 ) )
#    if face !=64206: 
#       print('Error gFXtfReadHead: MagicNumber !=FACE')
    fId.seek(8,1)
    HNumBytesThisRecord = int( fromfile( fId, int32, 1 ) )
    # Total byte count for this ping including this ping header. 
    # Isis Note: Isis records data packets in multiples of 64 bytes. 
    # If the size of the data packet is not an exact multiple of 64 bytes, 
    # zeros are padded at the end packet and this value will be promoted to the next 64-byte granularity. 
    # In all cases, this value will be the EXACT size of this packet.
    fId.seek(HNumBytesThisRecord-14,1)
# ===End Calc Num of ShortHeader structure

# ===Begin ShortHeader structure Allocate
#RFace               = asarray( [0]*nRec )
RHeaderType         = asarray( [0]*nRec )
RSubChannelNumber   = asarray( [0]*nRec )
RNumChansToFollow   = asarray( [0]*nRec )
RUnused             = asarray( [0]*nRec )
RNumBytesThisRecord = asarray( [0]*nRec )
RSeek               = asarray( [0]*nRec )
ROnFlag             = asarray( [0]*nRec )
# ===End ShortHeader structure Allocate

fId.seek(ss,0)

for m in range(nRec):
    #===Begin ShortHeader structure Read
    RFace = int( fromfile( fId, int16, 1 ) ) 
    #if XtfRFace~=64206, error('Error gFXtfReadHead: MagicNumber~=FACE');end
    RHeaderType[m]       = int( fromfile( fId, int8, 1 ) )
    RSubChannelNumber[m] = int( fromfile( fId, int8, 1 ) ) 
    # If HeaderType is bathymetry, this indicates which head; if HeaderType is forward-looking sonar, 
    # and then this indicates which array. Also, Klein 5000 beam numbers are logged here. 
    RNumChansToFollow[  m] = int( fromfile( fId, int16, 1 ) ) # If HeaderType is sonar, number of channels to follow.
    RUnused[            m] = int( fromfile( fId, int32, 1 ) ) # Unused. Set to 0
    RNumBytesThisRecord[m] = int( fromfile( fId, int32, 1 ) )
    # Total byte count for this ping including this ping header. 
    # Isis Note: Isis records data packets in multiples of 64 bytes.
    # If the size of the data packet is not an exact multiple of 64 bytes, 
    # zeros are padded at the end packet and this value will be promoted to the next 64-byte granularity. 
    # In all cases, this value will be the EXACT size of this packet.
    ROnFlag[m] = 1
    RSeek[  m] = fId.tell()
    fId.seek(RNumBytesThisRecord[m]-14,1)
    #===End ShortHeader structure Read

fId.close()

SubCh = 0
LHead = asarray( [0]*nRec )

fId = open(fName, "rb")

for m in range(nRec):
    if (RHeaderType[m]==0)&(RSubChannelNumber[m]==SubCh):
       LHead[m] = 1

LenHead = sum(LHead)

nHead = asarray( [0]*LenHead )
i     = 0

for m in range(nRec):
    if LHead[m] == 1:
       nHead[i] = m
       i = i + 1

ChF = max( RNumChansToFollow[nHead] )

#===Begin Header and ChanInfo Allocate

HMessageNum    = zeros(LenHead)
HYear          = zeros(LenHead)
HMonth         = zeros(LenHead)
HDay           = zeros(LenHead)
HHour          = zeros(LenHead)
HMinute        = zeros(LenHead)
HSecond        = zeros(LenHead)
HHSeconds      = zeros(LenHead)
HJulianDay     = zeros(LenHead)
HEventNumber   = zeros(LenHead)
HPingNumber    = zeros(LenHead)
HSoundVelocity = zeros(LenHead)
HOceanTide     = zeros(LenHead)
HReserved2     = zeros(LenHead)
HConductivityFreq = zeros(LenHead)
HTemperatureFreq  = zeros(LenHead)
HPressureFreq     = zeros(LenHead) 
HPressureTemp     = zeros(LenHead)
HConductivity     = zeros(LenHead)
HWaterTemperature = zeros(LenHead)
HPressure         = zeros(LenHead)
HComputedSoundVelocity = zeros(LenHead)
HMagX = zeros(LenHead)
HMagY = zeros(LenHead)
HMagZ = zeros(LenHead)
HAuxVal1 = zeros(LenHead)
HAuxVal2 = zeros(LenHead)
HAuxVal3 = zeros(LenHead)
HAuxVal4 = zeros(LenHead)
HAuxVal5 = zeros(LenHead)
HAuxVal6 = zeros(LenHead)
HSpeedLog  = zeros(LenHead)
HTurbidity = zeros(LenHead)
HShipSpeed = zeros(LenHead)
HShipGyro  = zeros(LenHead)
HShipYcoordinate = zeros(LenHead)
HShipXcoordinate = zeros(LenHead)
HShipAltitude    = zeros(LenHead)
HShipDepth       = zeros(LenHead)
HFixTimeHour     = zeros(LenHead)
HFixTimeMinute   = zeros(LenHead)
HFixTimeSecond   = zeros(LenHead)
HFixTimeHsecond  = zeros(LenHead)
HSensorSpeed     = zeros(LenHead)
HKP              = zeros(LenHead)
HSensorYcoordinate = zeros(LenHead)
HSensorXcoordinate = zeros(LenHead)
HSonarStatus       = zeros(LenHead)
HRangeToFish       = zeros(LenHead)
HBearingToFish     = zeros(LenHead)
HCableOut          = zeros(LenHead)
HLayback           = zeros(LenHead)
HCableTension      = zeros(LenHead)
HSensorDepth       = zeros(LenHead)
HSensorPrimaryAltitude = zeros(LenHead)
HSensorAuxAltitude     = zeros(LenHead)
HSensorPitch           = zeros(LenHead)
HSensorRoll            = zeros(LenHead)
HSensorHeading         = zeros(LenHead)
HHeave                 = zeros(LenHead)
HYaw                   = zeros(LenHead)
HAttitudeTimeTag       = zeros(LenHead)
HDOT                   = zeros(LenHead)
HNavFixMilliseconds    = zeros(LenHead)
HComputerClockHour     = zeros(LenHead)
HComputerClockMinute   = zeros(LenHead)
HComputerClockSecond   = zeros(LenHead)
HComputerClockHsec     = zeros(LenHead)
HFishPositionDeltaX    = zeros(LenHead)
HFishPositionDeltaY    = zeros(LenHead)
HFishPositionErrorCode = chararray(LenHead)
HOptionalOffsey        = zeros(LenHead)
HCableOutHundredths    = zeros(LenHead)
HReservedSpace2        = zeros((6,LenHead))
CChannelNumber         = zeros((ChF,LenHead))
CDownsampleMethod      = zeros((ChF,LenHead))
CSlantRange            = zeros((ChF,LenHead))
CGroundRange           = zeros((ChF,LenHead))
CTimeDelay             = zeros((ChF,LenHead))
CTimeDuration          = zeros((ChF,LenHead))
CSecondsPerPing        = zeros((ChF,LenHead))
CProcessingFlags       = zeros((ChF,LenHead))
CFrequency             = zeros((ChF,LenHead))
CInitialGainCode       = zeros((ChF,LenHead))
CGainCode              = zeros((ChF,LenHead))
CBandWidth             = zeros((ChF,LenHead))
CContactNumber         = zeros((ChF,LenHead))
CContactClassification = zeros((ChF,LenHead))
CContactSubNumber      = zeros((ChF,LenHead))
CContactType           = zeros((ChF,LenHead))
CNumSamples            = zeros((ChF,LenHead))
CMillivoltScale        = zeros((ChF,LenHead))
CContactTimeOffTrack   = zeros((ChF,LenHead))
CContactCloseNumber    = zeros((ChF,LenHead))
CReserved2             = zeros((ChF,LenHead))
CFixedVSOP             = zeros((ChF,LenHead))
CWeight                = zeros((ChF,LenHead))
CReservedSpace         = zeros((ChF,LenHead,4))

#===End Header and ChanInfo Allocate

df = 0
fId.seek(0,0)

CDataSeek = zeros((ChF,LenHead))

for n in range(LenHead):
    fId.seek(RSeek[nHead[n]]-df,1)
    #===Begin Header Read
    HMessageNum[n] = nHead[n]
    HYear[       n] = int( fromfile( fId, int16, 1 ) ) # Ping year
    HMonth[      n] = int( fromfile( fId, int8 , 1 ) ) # Ping month
    HDay[        n] = int( fromfile( fId, int8 , 1 ) ) # Ping day
    HHour[       n] = int( fromfile( fId, int8 , 1 ) ) # Ping hour
    HMinute[     n] = int( fromfile( fId, int8 , 1 ) ) # Ping minute
    HSecond[     n] = int( fromfile( fId, int8 , 1 ) ) # Ping seconds
    HHSeconds[   n] = int( fromfile( fId, int8 , 1 ) ) # Ping hundredths of seconds (0-99)
    HJulianDay[  n] = int( fromfile( fId, int16, 1 ) ) # Julian day of a pings occurrence.
    HEventNumber[n] = int( fromfile( fId, int32, 1 ) ) # Last logged event number; nav interface template token=O. 
    # NOTE: In Isis v4.30 and earlier EventNumber field was located at byte 26-27 and was a two byte WORD.  
    # At byte 24-25 there used to be a WORD CurrentLineID.  
    # The CurrentLineID field no longer exists in the .XTF format.  
    # Therefore, to read the event number correctly an application MUST check the Isis version string 
    # starting at byte 10 of the XTFFILEHEADER structure.
    HPingNumber[n] = int( fromfile( fId, int32, 1 ) ) # Counts consecutively (usually from 0) and increments for each update.  
    # Isis Note: The counters are different between sonar and bathymetry updates.
    HSoundVelocity[n] = float( fromfile( fId, float32, 1 ) )  # m/s,  
    # Isis uses 750 (one way), some XTF files use 1500.  
    # NOTE: Can be changed on Isis menu.  
    # This value is never computed and can only be changed manually by the user. See ComputedSoundVelocity below.
    HOceanTide[n] = float( fromfile( fId, float32, 1 ) )  # Altitude above Geoide (from RTK), 
    # if present; ELSE Ocean tide in meters; nav interface template token = {t} 
    # Isis Note: Can be changed by the user on the Configure menu in Isis.
    HReserved2[n] = int( fromfile( fId, int32  , 1 ) ) # Unused. Set to 0.
    HConductivityFreq[n] = float( fromfile( fId, float32, 1 ) )  # Conductivity frequency in Hz. 
    # nav interface template token = Q Raw CTD information.  
    # The Freq values are those sent up by the Seabird CTD. The Falmouth Scientific CTD sends up computed data.
    HTemperatureFreq[n] = float( fromfile( fId, float32, 1 ) )  # Temperature frequency in Hz. 
    # nav interface template token = b Raw CTD information.  
    # The Freq values are those sent up by the Seabird CTD. The Falmouth Scientific CTD sends up computed data.
    HPressureFreq[n] = float( fromfile( fId, float32, 1 ) )  # Pressure frequency in Hz. nav interface template token = 0. 
    # Raw CTD information.  
    # The Freq values are those sent up by the Seabird CTD. The Falmouth Scientific CTD sends up computed data.
    HPressureTemp[n] = float( fromfile( fId, float32, 1 ) )  # Pressure temperature (Degrees C); 
    # nav interface template token = ; Raw CTD information.  
    # The Freq values are those sent up by the Seabird CTD. The Falmouth Scientific CTD sends up computed data.
    HConductivity[n] = float( fromfile( fId, float32, 1 ) )  # Conductivity in Siemens/m; 
    # nav interface token = {c}; can be computed from Q Computed CTD information. 
    # When using a Seabird CTD, these values are computed from the raw Freq values (above).
    HWaterTemperature[n] = float( fromfile( fId, float32, 1 ) )  # Water temperature in Celsius. nav interface token = {w}; 
    # can be computed from b. 
    # Computed CTD information. When using a Seabird CTD, these values are computed from the raw Freq values (above)
    HPressure[n] = float( fromfile( fId, float32, 1 ) )  # Water pressure in psia; 
    # nav interface token = {p}; can be computed from 0. Computed CTD information. 
    # When using a Seabird CTD, these values are computed from the raw Freq values (above).
    HComputedSoundVelocity[n]=float( fromfile( fId, float32, 1 ) )  # Meters/second computed from 
    # Conductivity, WaterTemperature, and Pressure using the Chen Millero formula (1977), formula (JASA, 62, 1129-1135)
    HMagX[n]    = float( fromfile( fId, float32, 1 ) )  # X-axis magnetometer data in mgauss. 
    # Nav interface template token = e. Sensors Information.
    HMagY[n]    = float( fromfile( fId, float32, 1 ) )  # Y-axis magnetometer data in mgauss. 
    # Nav interface template token = w. Sensors Information.
    HMagZ[n]    = float( fromfile( fId, float32, 1 ) )  # Z-axis magnetometer data in mgauss. 
    # Nav interface template token = z. Sensors Information.
    HAuxVal1[n] = float( fromfile( fId, float32, 1 ) )  # Sensors Information. Nav interface template token = 1. 
    # Auxiliary values can be used to store and display any value at the user's discretion. Not used in any calculation in Isis or Target. 
    # Isis Note: Displayed in the Sensors window by selecting Window?Text?Sensors
    HAuxVal2[n] = float( fromfile( fId, float32, 1 ) )  # Sensors Information. Nav interface template token = 2. 
    # Auxiliary values can be used to store and display any value at the user's discretion.  
    # These are not used in any calculation in Isis or Target.
    # Isis Note: Displayed in the Sensors window by selecting Window?Text?Sensors
    HAuxVal3[n] = float( fromfile( fId, float32, 1 ) )  # Sensors Information. Nav interface template token = 3. 
    # Auxiliary values can be used to store and display any value at the user's discretion.   
    # These are not used in any calculation in Isis or Target. 
    # Isis Note: Displayed in the Sensors window by selecting Window?Text?Sensors
    HAuxVal4[n] = float( fromfile( fId, float32, 1 ) )  # Sensors Information. Nav interface template token = 4. 
    # Auxiliary values can be used to store and display any value at the user's discretion. 
    # These are not used in any calculation in Isis or Target. 
    # Isis Note: Displayed in the Sensors window by selecting Window?Text?Sensors
    HAuxVal5[n] = float( fromfile( fId, float32, 1 ) )  # Sensors Information. Nav interface template token = 5. 
    # Auxiliary values can be used to store and display any value at the user's discretion. 
    # These are not used in any calculation in Isis or Target. 
    # Isis Note: Displayed in the Sensors window by selecting Window?Text?Sensors
    HAuxVal6[n] = float( fromfile( fId, float32, 1 ) )  # Sensors Information. Nav interface template token = 6. 
    # Auxiliary values can be used to store and display any value at the user's discretion.
    # These are not used in any calculation in Isis or Target. Isis Note: Displayed in the Sensors window by selecting Window?Text?Sensors
    HSpeedLog[n] = float( fromfile( fId, float32, 1 ) )  # Sensors Information. Speed log sensor on towfish in knots.
    # NOTE: This is not fish speed. Nav interface template token = s.
    HTurbidity[n] = float( fromfile( fId, float32, 1 ) )  # Sensors Information. 
    # Turbidity sensor (0 to +5 volts) multiplied by 10000. nav interface template token = | (the pipe symbol).
    HShipSpeed[n] = float( fromfile( fId, float32, 1 ) )  # Ship Navigation information. Ship speed in knots. 
    # nav interface template token = v. 
    # Isis Note: These values are stored only and are not part of any equation or computation in Isis.
    HShipGyro[n] = float( fromfile( fId, float32, 1 ) )  # Ship Navigation information. Ship gyro in degrees. 
    # nav interface template token = G. 
    # Isis Note: This is used as the directional sensor for Multibeam Bathymetry data.
    HShipYcoordinate[n] = float( fromfile( fId, float64, 1 ) ) # Ship Navigation information. 
    # Ship latitude or northing in degrees. nav interface template token = y. 
    # Isis Note: These values are stored only and are not part of any equation or computation in Isis.
    HShipXcoordinate[n] = float( fromfile( fId, float64, 1 ) ) # Ship Navigation information. 
    # Ship longitude or easting in degrees. nav interface template token = x.
    # Isis Note: These values are stored only and are not part of any equation or computation in Isis.
    HShipAltitude[n] = int( fromfile( fId, int16, 1 ) ) # Ship altitude in decimeters
    HShipDepth[   n] = int( fromfile( fId, int16, 1 ) ) # Ship depth in decimeters.
    HFixTimeHour[ n] = int( fromfile( fId, int8 , 1 ) ) # Sensor Navigation information.
    # Hour of most recent nav update. nav interface template token = H. 
    # Isis Note: The time of the nav is adjusted by the NavLatency stored in the XTF file header.
    HFixTimeMinute[n] = int( fromfile( fId, int8, 1 ) ) # Sensor Navigation information.
    # Minute of most recent nav update. nav interface template token = I. 
    # Isis Note: The time of the nav is adjusted by the NavLatency stored in the XTF file header.
    HFixTimeSecond[n] = int( fromfile( fId, int8, 1 ) ) # Sensor Navigation information.
    # Second of most recent nav update. nav interface template token = S. 
    # Isis Note: The time of the nav is adjusted by the NavLatency stored in the XTF file header.
    HFixTimeHsecond[n] = int( fromfile( fId, int8, 1 ) ) # Sensor Navigation information. Hundredth of a Second of most recent nav update.
    # Isis Note: The time of the nav is adjusted by the NavLatency stored in the XTF file header.
    HSensorSpeed[n] = float( fromfile( fId, float32, 1 ) )  # Sensor Navigation information. Speed of towfish in knots.
    # Used for speed correction and position calculation; nav interface template token = V.
    HKP[n] = float( fromfile( fId, float32, 1 ) )  #Sensor Navigation information. Kilometers Pipe; nav interface template token = {K}.
    HSensorYcoordinate[n] = float( fromfile( fId, float64, 1 ) ) # Sensor Navigation information.
    # Sensor latitude or northing; nav interface template token = E. 
    # NOTE: when NavUnits in the file header is 0, values are in meters (northings and eastings).  
    # When NavUnits is 3, values are in Lat/Long.  Also see the Layback value, below.
    HSensorXcoordinate[n] = float( fromfile( fId, float64, 1 ) ) # Sensor Navigation information.
    # Sensor longitude or easting; nav interface template token = N. 
    # NOTE: when NavUnits in the file header is 0, values are in meters (northings and eastings). 
    # When NavUnits is 3, values are in Lat/Long. Also see the Layback value, below.
    HSonarStatus[n] = int( fromfile( fId, int16  , 1 ) ) # Tow Cable information. 
    # System status value, sonar dependant (displayed in Status window).
    HRangeToFish[n] = int( fromfile( fId, int16, 1 ) )   # Slant range to sensor in decimeters; 
    # nav interface template token = ? (question mark). 
    # Stored only  not used in any computation.
    HBearingToFish[n] = int(   fromfile( fId, int16  , 1 ) ) # Bearing to towfish from ship, stored in degrees multiplied by 100; 
    # nav interface template token = > (greater-than sign]. 
    # Stored only  not used in any computation in Isis.
    HCableOut[n] = int(   fromfile( fId, int16  , 1 ) ) # Tow Cable information. Amount of cable payed out in meters; 
    # nav interface template token = o.
    HLayback[n] = float( fromfile( fId, float32, 1 ) )  # Tow Cable information.
    # Distance over ground from ship to fish.; nav interface template token = l. 
    # Isis Note: When this value is non-zero, 
    # Isis assumes that SensorYcoordinate and SensorXcoordinate need to be adjusted with the Layback. 
    # The sensor position is then computed using the current sensor heading and this layback value.
    # The result is displayed when a position is computed in Isis.
    HCableTension[n] = float( fromfile( fId, float32, 1 ) )  # Tow Cable information Cable tension from serial port. 
    # Stored only; nav interface template token = P
    HSensorDepth[n] = float( fromfile( fId, float32, 1 ) )  # Sensor Altitude information. Distance (m) from sea surface to sensor.
    # The deeper the sensor goes, the bigger (positive) this value becomes. nav interface template token = 0 (zero)
    HSensorPrimaryAltitude[n] = float( fromfile( fId, float32, 1 ) )  # Sensor Altitude information. 
    # Distance from towfish to the sea floor; nav interface template token = 7. 
    # Isis Note: This is the primary altitude as tracked by the Isis bottom tracker or entered manually by the user. 
    # Although not recommended, the user can override the Isis bottom tracker by sending the primary altitude over the serial port.
    # The user should turn the Isis bottom tracker Off when this is done.
    HSensorAuxAltitude[n] = float( fromfile( fId, float32, 1 ) )  # Sensor Altitude information. 
    # Auxiliary altitude; nav interface template token = a. 
    # Isis Note: This is an auxiliary altitude as transmitted by an altimeter and received over a serial port.
    # The user can switch between the Primary and  Aux altitudes via the "options" button in the Isis bottom track window.
    HSensorPitch[n] = float( fromfile( fId, float32, 1 ) )  # Sensor Altitude information. 
    # Pitch in degrees (positive=nose up); nav interface template token = 8.
    HSensorRoll[n] = float( fromfile( fId, float32, 1 ) )  # Sensor Altitude information. 
    # Roll in degrees (positive=roll to starboard); nav interface template token = 9.
    HSensorHeading[n] = float( fromfile( fId, float32, 1 ) )  # Sensor Altitude information. 
    # Sensor heading in degrees; nav interface template token = h.
    HHeave[n] = float( fromfile( fId, float32, 1 ) )  # Attitude information. 
    # Sensors heave at start of ping. Positive value means sensor moved up. 
    # NOTE: These Pitch, Roll, Heading, Heave and Yaw values are those received closest in time to this sonar or bathymetry update.  
    # If a TSS or MRU is being used with a multibeam/bathymetry sensor, 
    # the user should  use the higher-resolution attitude data found in the XTFATTITUDEDATA structures.
    HYaw[n] = float( fromfile( fId, float32, 1 ) )  # Attitude information. Sensor yaw. 
    # Positive means turn to right. 
    # Note: These Pitch, Roll, Heading, Heave and Yaw values are those received closest in time to this sonar or bathymetry update. 
    # If a TSS or MRU is being used with a multibeam/bathymetry sensor, the user should use the higher-resolution attitude data 
    # found in the XTFATTITUDEDATA structures. 
    # Since the heading information is updated in high resolution, 
    # it is not necessary to log or use Yaw in any processing.  Isis does not use Yaw
    HAttitudeTimeTag[n] = int( fromfile( fId, int32, 1 ) ) # Altitude information.
    # In milliseconds - used to coordinate with millisecond time value in Attitude packets. 
    # (M)andatory when logging XTFATTITUDE packets.
    HDOT[n] = float( fromfile( fId, float32, 1 ) )  # Misc. Distance Off Track
    HNavFixMilliseconds[n] = int( fromfile( fId, int32, 1 ) ) # Misc. millisecond clock value when nav received
    HComputerClockHour[ n] = int( fromfile( fId, int8, 1 ) ) # Isis Note: The Isis computer clock time when this ping was received. 
    # May be different from ping time at start of this record if the sonar time-stamped the data and the two systems aren't synched. 
    # This time should be ignored in most cases
    HComputerClockMinute[n] = int( fromfile( fId, int8, 1 ) ) # Isis Note: see above Isis Note
    HComputerClockSecond[n] = int( fromfile( fId, int8, 1 ) ) # Isis Note: see above Isis Note
    HComputerClockHsec[  n] = int( fromfile( fId, int8, 1 ) ) # Isis Note: see above Isis Note
    HFishPositionDeltaX[ n] = float( fromfile( fId, float16, 1 ) ) # Additional Tow Cable and Fish information from Trackpoint.
    # Stored as meters multiplied by 3.0, supporting +/- 10000.0m (usually from trackpoint); nav interface template token = {DX}.
    HFishPositionDeltaY[n] = int( fromfile( fId, int16, 1 ) ) # Additional Tow Cable and Fish information from Trackpoint. 
    # X, Y offsets can be used instead of logged layback.; nav interface template token = {DY}.
    HFishPositionErrorCode[n] = fId.read(1) # Additional Tow Cable and Fish information from Trackpoint.
    # Error code for FishPosition delta x,y. (typically reported by Trackpoint).
    HOptionalOffsey[n] = int( fromfile( fId, int32, 1 ) ) # OptionalOffsey (Triton 7125 only)
    HCableOutHundredths[n] = int( fromfile( fId, int8, 1 ) ) # Hundredths of a meter of cable out, to be added  to the CableOut field.
    HReservedSpace2[0,n] = int( fromfile( fId, int8, 1 ) ) # Unused. Set to 0.
    HReservedSpace2[1,n] = int( fromfile( fId, int8, 1 ) )
    HReservedSpace2[2,n] = int( fromfile( fId, int8, 1 ) )
    HReservedSpace2[3,n] = int( fromfile( fId, int8, 1 ) )
    HReservedSpace2[4,n] = int( fromfile( fId, int8, 1 ) )
    HReservedSpace2[5,n] = int( fromfile( fId, int8, 1 ) )
    #===End Header Read
    
    for nn in range(RNumChansToFollow[nHead[n]]):
    
        #===Begin ChanInfo Read
        
	CChannelNumber[nn,n] = int( fromfile( fId, int16, 1 ) ) 
	# Typically 0=port (low frequency) 1=stbd (low frequency) 2=port (high frequency) 3=stbd (high frequency)
        CDownsampleMethod[nn,n] = int( fromfile( fId, int16, 1 ) ) # 2 = MAX; 4 = RMS
        CSlantRange[ nn,n] = float( fromfile( fId, float32, 1 ) ) # Slant range of the data in meters
        CGroundRange[nn,n] = float( fromfile( fId, float32, 1 ) ) # Ground range of the data; in meters (SlantRange^2 - Altitude^2)
        CTimeDelay[  nn,n] = float( fromfile( fId, float32, 1 ) ) # Amount of time, in seconds, to the start of recorded data. 
        # (almost always 0.0).
        CTimeDuration[   nn,n] = float( fromfile( fId, float32, 1 ) ) # Amount of time, in seconds, recorded (typically SlantRange/750)
        CSecondsPerPing[ nn,n] = float( fromfile( fId, float32, 1 ) ) # Amount of time, in seconds, from ping to ping. (SlantRange/750)
        CProcessingFlags[nn,n] = int(   fromfile( fId, int16  , 1 ) ) # 4 = TVG; 8 = BAC&GAC; 16 = filter, etc. (almost always zero)
        CFrequency[      nn,n] = int(   fromfile( fId, int16  , 1 ) ) # Ccenter transmit frequency for this channel
        CInitialGainCode[nn,n] = int(   fromfile( fId, int16  , 1 ) ) # Settings as transmitted by sonar
        CGainCode[       nn,n] = int(   fromfile( fId, int16  , 1 ) ) # Settings as transmitted by sonar
        CBandWidth[      nn,n] = int(   fromfile( fId, int16  , 1 ) ) # Settings as transmitted by sonar
        CContactNumber[  nn,n] = int(   fromfile( fId, int32  , 1 ) ) # Contact information . 
        # Upated when contacts are saved in Target utility.
        CContactClassification[nn,n] = int( fromfile( fId, int16, 1 ) ) # Contact information . 
        # Updated when contacts are saved in Target utility.
        CContactSubNumber[nn,n] = int( fromfile( fId, int8, 1 ) ) # Contact information. 
        # Udated when contacts are saved in Target utility
        CContactType[nn,n] = int( fromfile( fId, int8 , 1 ) ) # Contact information . Updated when contacts are saved in Target utility
        CNumSamples[ nn,n] = int( fromfile( fId, int32, 1 ) ) # Number of samples that will follow this structure. 
	# The number of bytes will be this value multiplied by the number of bytes per sample. 
        # BytesPerSample found in CHANINFO structure (given in the file header).
        CMillivoltScale[nn,n] = int( fromfile( fId, int16, 1 ) ) # Maximum voltage, in mv, 
        # represented by a full-scale value in the data.
	# If zero, then the value stored in the VoltScale should be used instead.
        # VoltScale can be found in the XTF file header, ChanInfo structure. 
	# Note that VoltScale is specified in volts, while MillivoltScale is stored in millivolts.
        # This provides for a range of 65,536 volts to 65,535 volts.
        CContactTimeOffTrack[nn,n] = float( fromfile( fId, float32, 1 ) ) # Time off track to this contact (stored in milliseconds)
        CContactCloseNumber[ nn,n] = int(   fromfile( fId, int8   , 1 ) )
        CReserved2[          nn,n] = int(   fromfile( fId, int8   , 1 ) ) # Unused. Set to 0.
        CFixedVSOP[          nn,n] = float( fromfile( fId, float32, 1 ) ) # This is the fixed, along-track size of each ping, 
        # stored in centimeters.
	# On multibeam systems with zero beam spread, this value needs to be filled in to prevent Isis from calculating  
        # along-track ground coverage based on beam spread and speed over ground.
        CWeight[nn,n] = int( fromfile( fId, int16, 1 ) ) # Weighting factor passed by some sonars,
	# this value is mandatory for Edgetech digital sonars types 24, 35, 38, 48 and Kongsberg SA type 48
        CReservedSpace[nn,n,0] = int( fromfile( fId, int8, 1 ) ) # Unused. Set to 0.
        CReservedSpace[nn,n,1] = int( fromfile( fId, int8, 1 ) )
        CReservedSpace[nn,n,2] = int( fromfile( fId, int8, 1 ) )
        CReservedSpace[nn,n,3] = int( fromfile( fId, int8, 1 ) )
        #===End ChanInfo Read
	
        #===Begin Data pass
        CDataSeek[nn,n] = fId.tell()
        fId.seek(CNumSamples[nn,n]*CBytesPerSample[nn],1)
        #===End Data pass
    df = fId.tell()

CNumSamples = int_(CNumSamples)
CDataSeek   = int_(CDataSeek  )
CUniPolar   = int_(CUniPolar  )
CReserved2  = int_(CReserved2 )
CContactType    = int_(CContactType)
CBytesPerSample = int_(CBytesPerSample)
CReservedSpace  = int_(CReservedSpace )
CContactCloseNumber = int_(CContactCloseNumber)

#===Begin Data Allocate
#Data = int_(zeros( (amax(CNumSamples),LenHead,ChF) ))
Data = zeros( (amax(CNumSamples),LenHead,ChF) )
#===End Data Allocate

#===Begin Data Read

#Descript_BytesPerSample_C = array([['int8','int16','int32'],['uint8','uint16','uint32']])
#Descript_UniPolar_Code    = array([0,1])
#Descript_BytesPerSample_Code = array([1,2,4]) 

df = 0
fId.seek(0,0)

#This is different from the Matlab code, since it expects that the data over all channels are stored with the same number of bytes... 

#for n in range(LenHead):
#    for nn in range(RNumChansToFollow[nHead[n]]):
#        fId.seek(CDataSeek[nn,n]-df,1)
#        for nnn in range(CNumSamples[nn,n]):
#            Data[nnn,n,nn] = int( fromfile( fId, int16, 1 ) )
#        df = fId.tell()

if CBytesPerSample[0] == 1:
   for n in range(LenHead):
       for nn in range(RNumChansToFollow[nHead[n]]):
           fId.seek(CDataSeek[nn,n]-df,1)
           for nnn in range(CNumSamples[nn,n]):
               Data[nnn,n,nn] = int( fromfile( fId, int8, 1 ) )
           df = fId.tell()
elif CBytesPerSample[0] == 2:
   for n in range(LenHead):
       for nn in range(RNumChansToFollow[nHead[n]]):
           fId.seek(CDataSeek[nn,n]-df,1)
           for nnn in range(CNumSamples[nn,n]):
               Data[nnn,n,nn] = int( fromfile( fId, int16, 1 ) )
           df = fId.tell()
else:
   for n in range(LenHead):
       for nn in range(RNumChansToFollow[nHead[n]]):
           fId.seek(CDataSeek[nn,n]-df,1)
           for nnn in range(CNumSamples[nn,n]):
               Data[nnn,n,nn] = int( fromfile( fId, int32, 1 ) )
           df = fId.tell()

#===End Data Read

fId.close()

print('Plotting...')

figure(1)
pcolor( abs(squeeze(Data[:,:,1])), vmin=0,vmax=1000)
colorbar()

show()

print("done.")
